import subprocess
import sys
from io import StringIO
from pathlib import Path
from typing import Optional, Set

import click
import paramiko  # type: ignore
from click.testing import CliRunner
from paramiko import SSHClient
from paramiko.sftp_client import SFTPClient  # type: ignore

from burny_common.copy_file_to_server import copy_file_to_server_helper, generate_path


def copy_folder_to_server_helper(
    client: SSHClient,
    sftp: SFTPClient,
    path_source_root_folder: Path,
    path_target_root_folder: Path,
    file_or_folder: Path,
    allowed_files: Optional[Set] = None,
):
    source_path_relative_to_root = file_or_folder.relative_to(path_source_root_folder)
    if file_or_folder.is_file():
        if allowed_files is not None and file_or_folder.absolute() not in allowed_files:
            # print(f"Ignoring file/folder {file_or_folder}")
            return
        # print(f"Copying file {file_or_folder}")
        path_target = path_target_root_folder / source_path_relative_to_root
        copy_file_to_server_helper(client, sftp, file_or_folder, path_target, create_target_folder=True)
    else:
        for file_or_folder2 in file_or_folder.iterdir():
            # Recursively copy folders
            copy_folder_to_server_helper(
                client,
                sftp,
                path_source_root_folder,
                path_target_root_folder,
                file_or_folder2,
                allowed_files=allowed_files,
            )


@click.command()
@click.option('--host', default='', help='host address')
@click.option('--port', default=22, help='port')
@click.option('--username', default='', help='user name')
@click.option('--password', default='', help='user password')
@click.option('--pkey', default='', help='private key')
@click.option('--sourcepath', default='', help='source folder to copy')
@click.option('--targetpath', default='', help='which folder the output should be')
@click.option('--respectgitignore', default=True, help='ignore files that are ignored by .gitignore')
def copy_folder_to_server(
    host: str,
    port: int,
    username: str,
    password: str,
    pkey: str,
    sourcepath: str,
    targetpath: str,
    respectgitignore: bool = True
):
    client: SSHClient
    with SSHClient() as client:
        client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        pkey_loaded = paramiko.RSAKey.from_private_key(StringIO(pkey))
        client.connect(hostname=host, port=port, username=username, password=password, pkey=pkey_loaded)

        path_source_root_folder = Path(sourcepath)
        path_target_root_folder = generate_path(client, targetpath)

        allowed_files: Optional[Set] = None
        # pylint: disable=R1732
        if respectgitignore:
            proc = subprocess.Popen(
                ['git', 'ls-files'],
                cwd=path_source_root_folder.absolute(),
                stdout=subprocess.PIPE,
            )
            output = proc.stdout
            if output is not None:
                files = output.read().decode()
                allowed_files = {(path_source_root_folder / Path(file)).absolute() for file in files.splitlines()}

        with client.open_sftp() as sftp:
            assert path_source_root_folder.is_dir()
            copy_folder_to_server_helper(
                client,
                sftp,
                path_source_root_folder,
                path_target_root_folder,
                # Start copying this folder
                path_source_root_folder,
                allowed_files=allowed_files,
            )
    sys.exit(0)


def main():
    runner = CliRunner()
    ip = 'some.url'
    username = 'some_name'
    key = 'my ssh key'
    result = runner.invoke(
        copy_folder_to_server, [
            f'--host={ip}',
            f'--username={username}',
            f'--pkey={key}',
            '--sourcepath=/home/burny/github/python-template',
            '--targetpath=test5',
        ]
    )
    for line in result.output.splitlines():
        print(line)
    if result.exit_code != 0:
        print(result.exception)
        print(result.exit_code)


if __name__ == '__main__':
    main()



import time
from contextlib import contextmanager

from loguru import logger


@contextmanager
def time_this(label: str):
    start = time.perf_counter_ns()
    yield
    end = time.perf_counter_ns()
    logger.info(f'TIME {label}: {(end - start) / 1e9} sec')


# Use like this
if __name__ == '__main__':
    with time_this('square rooting'):
        for i in range(10**4):
            _x = i**0.5



import os
import signal
import socket
import subprocess
import time
from contextlib import suppress
from pathlib import Path
from typing import Optional, Set

import psutil
import requests
from loguru import logger

WEBSITE_IP = 'http://localhost'


class Timeout:
    """
    Run something for a maximum limited time
    try:
        with Timeout(seconds=2):
            ...
    except TimeoutError:
        ...
    """

    def __init__(self, seconds=1, error_message='Timeout'):
        self.seconds = seconds
        self.error_message = error_message

    def handle_timeout(self, signum, frame):
        raise TimeoutError(self.error_message)

    def __enter__(self):
        signal.signal(signal.SIGALRM, self.handle_timeout)
        signal.alarm(self.seconds)

    def __exit__(self, type_, value, traceback):
        signal.alarm(0)


def get_pid(name: str) -> Set[int]:
    """ Return a list of PIDs of all processes with the exact given name. """
    process_pids = set()
    for proc in psutil.process_iter():
        if name == proc.name():
            pid = proc.pid
            process_pids.add(pid)
    return process_pids


def remove_leftover_files(files: Set[Path]):
    for file in files:
        if file.is_file():
            os.remove(file)


def is_port_free(port: int) -> bool:
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        sock.bind(('', port))
        sock.close()
        return True
    except OSError:
        return False


def find_next_free_port(port: int = 10_000, max_port: int = 65_535, exclude_ports: Optional[Set[int]] = None) -> int:
    if exclude_ports is None:
        exclude_ports = set()

    while port <= max_port:
        if port not in exclude_ports and is_port_free(port):
            return port
        port += 1
    raise IOError('No free ports')


# pylint: disable=R1732
def check_if_docker_is_running() -> bool:
    p = subprocess.Popen(['docker', 'ps'], stdout=subprocess.PIPE)
    _return_code = p.wait()
    if not p.stdout:
        return False
    output = p.stdout.read().decode()
    docker_running = output.startswith('CONTAINER ID')
    if docker_running:
        logger.info('Docker running detected')
    return docker_running


def get_website_address(port: int) -> str:
    return f'{WEBSITE_IP}:{port}'


def start_svelte_dev_server(
    port: int,
    newly_created_processes: Set[int],
    frontend_folder_path: Path,
    backend_proxy: str = 'localhost:8000',
):
    currently_running_node_processes = get_pid('node')

    assert is_port_free(port), f'Unable to start svelte dev server because port {port} is blocked'
    logger.info(f'Starting frontend on port {port}')
    _ = subprocess.Popen(
        ['npx', 'cross-env', f'BACKEND_SERVER={backend_proxy}', 'svelte-kit', 'dev', '--port', f'{port}'],
        cwd=frontend_folder_path,
        env=os.environ.copy()
    )
    # Give it some time to create dev server and all (3?) node proccesses
    with Timeout(10, 'Took more than 10 seconds'):
        while is_port_free(port):
            time.sleep(0.1)
        while 1:
            try:
                result = requests.get(get_website_address(port))
                if result.status_code == 200:
                    break
            except requests.exceptions.ConnectionError:
                pass
            time.sleep(0.1)

    new_processes = get_pid('node') - currently_running_node_processes
    logger.info(f'New node processes: {new_processes}')
    newly_created_processes |= new_processes


def start_fastapi_dev_server(
    port: int,
    newly_created_processes: Set[int],
    fastapi_folder_path: Path,
):
    print(str(fastapi_folder_path.absolute()))
    currently_running_uvicorn_processes = get_pid('uvicorn')

    # Why does this return errors even when fastapi server is not running
    # assert is_port_free(port), f"Unable to start fastapi server because port {port} is blocked"
    logger.info(f'Listing poetry packages')
    command_temp = ['poetry', 'show', '-v']
    _ = subprocess.Popen(
        command_temp,
        cwd=fastapi_folder_path,
        env=os.environ.copy(),
    )
    logger.info(f'Starting backend on port {port}')
    command = ['poetry', 'run', 'uvicorn', 'main:app', '--host', 'localhost', '--port', f'{port}']
    logger.info(f'In work directory "{fastapi_folder_path}" running command: {command}')
    _ = subprocess.Popen(
        command,
        cwd=fastapi_folder_path,
        env=os.environ.copy(),
    )
    # Give it some time to create backend dev server
    with Timeout(10, 'Took more than 10 seconds'):
        while is_port_free(port):
            time.sleep(0.1)
        while 1:
            with suppress(requests.exceptions.ConnectionError):
                result = requests.get(get_website_address(port))
                if result.status_code == 200:
                    break
            time.sleep(0.1)

    new_processes: Set[int] = get_pid('uvicorn') - currently_running_uvicorn_processes
    logger.info(f'New uvicorn processes: {new_processes}')
    newly_created_processes |= new_processes


# def check_if_mongodb_is_running(mongo_db_port: int = 27017) -> bool:
#     mongo_db_address = f'mongodb://localhost:{mongo_db_port}'
#     try:
#         with Timeout(seconds=2):
#             _my_client: MongoClient
#             with pymongo.MongoClient(mongo_db_address) as _my_client:
#                 pass
#     except TimeoutError:
#         return False
#     return True
#
#
# # pylint: disable=R1732
# def start_mongodb(mongo_db_port: int = 27017) -> int:
#     # Start mongodb via docker
#     if check_if_mongodb_is_running(mongo_db_port):
#         logger.info(f'MongoDB is already running on port {mongo_db_port}')
#         return mongo_db_port
#     command = [
#         # TODO add volume to save db
#         'docker',
#         'run',
#         '--rm',
#         '-d',
#         '--name',
#         'mongodb_test',
#         '-p',
#         # TODO use mongo_db_port
#         '27017-27019:27017-27019',
#         'mongo:5.0.0',
#     ]
#     logger.info(f"Starting mongoDB with command: {' '.join(command)}")
#     process = subprocess.Popen(command)
#     process.wait()
#     return mongo_db_port


def check_if_postgres_is_running(port: int = 5432) -> bool:
    # If we can connect to port 5432, postgres is already running
    # TODO find better way to figure out if postgress is running
    return not is_port_free(port)


# pylint: disable=R1732
def start_postgres(postgres_port: int = 5432) -> int:
    # Start postgres via docker
    if check_if_postgres_is_running(postgres_port):
        logger.info(f'Postgres is already running on port {postgres_port}')
        return postgres_port
    postgres_container_name = 'postgres_test'
    postgres_volume_name = 'postgres_test'
    postgres_username = 'postgres'
    postgres_password = 'changeme'
    postgres_image = 'postgres:9.6.23-alpine3.14'
    # postgres_port = find_next_free_port()
    if check_if_docker_is_running():
        # docker run --rm --name postgres_test -p 5432:5432 --volume postgres_test:/data/postgres -e POSTGRES_USER=postgres -e POSTGRES_PASSWORD=changeme postgres:9.6.23-alpine3.14
        command = [
            # TODO add volume to save db, or should that not be active while testing?
            'docker',
            'run',
            '--rm',
            '-d',
            '--name',
            postgres_container_name,
            '-p',
            f'{postgres_port}:{postgres_port}',
            '--volume',
            f'{postgres_volume_name}:/data/postgres',
            '-e',
            f'POSTGRES_USER={postgres_username}',
            '-e',
            f'POSTGRES_PASSWORD={postgres_password}',
            postgres_image,
        ]
        logger.info(f"Starting postgres with command: {' '.join(command)}")
        _process = subprocess.Popen(command)
    else:
        raise NotImplementedError()
    return postgres_port


def kill_processes(processes: Set[int]):
    # Soft kill
    for pid in processes:
        logger.info(f'Killing {pid}')
        with suppress(ProcessLookupError):
            os.kill(pid, signal.SIGTERM)
    time.sleep(.1)

    # Force kill
    for pid in processes:
        logger.info(f'Force killing {pid}')
        with suppress(ProcessLookupError):
            os.kill(pid, signal.SIGKILL)


if __name__ == '__main__':
    free_frontend_port = find_next_free_port()
    free_backend_port = find_next_free_port(exclude_ports={free_frontend_port})
    start_fastapi_dev_server(free_backend_port, set(), Path(__file__).parents[2] / "fastapi_server")
    start_svelte_dev_server(
        free_frontend_port,
        set(),
        Path(__file__).parents[2] / "svelte_frontend",
        backend_proxy=f'localhost:{free_backend_port}'
    )
    logger.info(f'Docker running: {check_if_docker_is_running()}')
    logger.info(f'Postgres running: {check_if_postgres_is_running()}')
    # logger.info(f'MongoDB running: {check_if_mongodb_is_running()}')
    # start_postgres()
    # start_mongodb()
    while 1:
        time.sleep(1)



import sys
from io import StringIO

import click
import paramiko  # type: ignore
from click.testing import CliRunner
from paramiko import SSHClient


@click.command()
@click.option('--host', default='', help='host address')
@click.option('--port', default=22, help='port')
@click.option('--username', default='', help='user name')
@click.option('--password', default='', help='user password')
@click.option('--pkey', default='', help='private key')
@click.option('--command', default='ls -la', help='command to execute')
def run_command_on_server(host: str, port: int, username: str, password: str, pkey: str, command: str):
    client: SSHClient
    with SSHClient() as client:
        client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        pkey_loaded = paramiko.RSAKey.from_private_key(StringIO(pkey))
        client.connect(hostname=host, port=port, username=username, password=password, pkey=pkey_loaded)
        _stdin, stdout, _stderr = client.exec_command(f'{command}')
        lines = stdout.readlines()
        for line in lines:
            print(line)
    sys.exit(0)


def main():
    runner = CliRunner()
    ip = 'some.url'
    username = 'some_name'
    key = 'my ssh key'
    result = runner.invoke(
        run_command_on_server, [
            f'--host={ip}',
            f'--username={username}',
            f'--pkey={key}',
            '--command=ls -lah',
        ]
    )
    for line in result.output.splitlines():
        print(line)
    if result.exit_code != 0:
        print(result.exception)
        print(result.exit_code)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['burny_common']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.1.3,<9.0.0',
 'loguru>=0.6.0,<0.7.0',
 'paramiko>=2.11.0,<3.0.0',
 'psutil>=5.9.1,<6.0.0',
 'requests>=2.28.0,<3.0.0',
 'types-requests>=2.27.31,<3.0.0']

setup_kwargs = {
    'name': 'burny-common',
    'version': '0.0.6',
    'description': '',
    'long_description': None,
    'author': 'BurnySc2',
    'author_email': 'gamingburny@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)
